from argparse import ArgumentParser
from collections import defaultdict
from itertools import product
from pathlib import Path
import typing as T

from numpy import nan
from pandas import DataFrame, merge, MultiIndex
import yaml

from parser import parse_mother_request
from grib import read_grib_params


PORTFOLIO_PATH = Path(__file__).resolve().parent.parent / "data_portfolio"
PORTFOLIO_KINDS = ["production", "reduced"]
PORTFOLIO_VERSIONS = [path.name for path in PORTFOLIO_PATH.iterdir()]
GENERATIONS = {
    1: "v0.1.1",
    2: "v1.2.0",
}
OUTPUT_PATH = Path(__file__).resolve().parent.parent / "variables"

HTML_COLORS = {
    "new_dgov": "#a4f3a4",
    # "new_dgov_odd": "#d3ffd3",
    "old_dgov": "#f3a4a4",
    # "old_dgov_odd": "#ffd3d3",
}
HTML_PATH = Path(__file__).resolve().parent.parent / "html"
HTML_FRAMEWORK = HTML_PATH / "framework.html"

DFHTML = []


apps = None
portfoilios = None
grib = None
compared = None
compared_apps = None
compared_mr = None


def _get_parser() -> ArgumentParser:
    parser = ArgumentParser()
    parser.add_argument(
        "path",
        help="Path for mother request",
        type=str,
    )
    return parser


def get_variables(path: Path | str, must_match: bool):
    df = parse_mother_request(
        path, ["gsv_param", "opa_variable", "opa_stat", "opa_stat_freq"]
    )
    if not (df["gsv_param"] == df["opa_variable"]).all():
        if must_match:
            raise ValueError("Not all GSV and OPA variables match.")
        raise NotImplementedError(
            "Not all GSV and OPA variables match. Functionality not implemented."
        )
    else:
        df.drop("gsv_param", axis=1, inplace=True)
        df.rename(columns={"opa_variable": "Variable"}, inplace=True)
        df.rename(columns={"opa_stat": "OPA Stat"}, inplace=True)
        df.rename(columns={"opa_stat_freq": "OPA StatFreq"}, inplace=True)
    return df


def read_data_portfolio(version: str, kind: str):
    if version not in PORTFOLIO_VERSIONS:
        raise ValueError(
            f"Version '{version}' is not valid. "
            f"Please, pass one of '{','.join(PORTFOLIO_VERSIONS)}'."
        )
    if kind not in PORTFOLIO_KINDS:
        raise ValueError(
            f"Kind '{kind}' is not valid. "
            f"Please, pass one of '{','.join(PORTFOLIO_KINDS)}'."
        )
    portfolio_path = PORTFOLIO_PATH / version / kind / "portfolio.yaml"
    if not portfolio_path.exists():
        raise FileNotFoundError(f"Malformed portfolio: {version}.")
    with open(portfolio_path, "r", encoding="utf-8") as rf:
        return yaml.safe_load(rf)


def get_portfolios():
    portfolios = {}
    for kind, (generation, version) in product(PORTFOLIO_KINDS, GENERATIONS.items()):
        portfolios[(kind, generation)] = read_data_portfolio(version=version, kind=kind)
    return portfolios


def find_matches(grib: DataFrame, variables: T.Sequence[str]):
    return {variable: grib[grib["Short name"] == variable] for variable in variables}


def _record_to_dct(record, models=None):
    variable_dct = record.to_dict(orient="records")[0]
    format_entry = variable_dct.pop("Format", None)
    if format_entry:
        variable_dct.update(
            {
                "GRIB1": "GRIB1" in format_entry,
                "GRIB2": "GRIB2" in format_entry,
            }
        )
    if models:
        variable_dct.update(
            {
                "IFS-FESOM": "ifs-fesom" in models,
                "IFS-NEMO": "ifs-nemo" in models,
                "ICON": "icon" in models,
            }
        )
    return variable_dct


def convert_variables(portfolio, grib: DataFrame):
    variables = {}
    variable_models = defaultdict(set)
    for model, model_entries in portfolio.items():
        for entry in model_entries:
            for variable in entry["variables"]:
                variable_models[variable].add(model)
    for model, model_entries in portfolio.items():
        for entry in model_entries:
            for variable in entry["variables"]:
                if variable not in variables:
                    variable_record = grib[grib["ID"] == variable]
                    variables[variable] = _record_to_dct(
                        variable_record, variable_models[variable]
                    )
    return DataFrame.from_records(list(variables.values()))


def dump_variables(portfolios, grib):
    for key, portfolio in portfolios.items():
        file_name = f"gen{key[1]}_{key[0]}.csv"
        variables_df = convert_variables(portfolio, grib)
        variables_df.to_csv(OUTPUT_PATH / file_name)


def compare_portfolios(df1, df2):
    on_cols = ["ID", "Name", "Short name", "Units", "GRIB1", "GRIB2"]
    df = merge(df1, df2, on=on_cols, how="outer", suffixes=("_1", "_2"), indicator=True)
    df["DGOV1"] = df["_merge"].map(lambda x: x in ("left_only", "both"))
    df["DGOV2"] = df["_merge"].map(lambda x: x in ("right_only", "both"))
    df.drop(
        ["Access method_1", "Access method_2", "GRIB1", "GRIB2", "_merge"],
        axis=1,
        inplace=True,
    )
    new_cols = [
        "ID",
        "Name",
        "Short name",
        "Units",
        "DGOV1",
        "IFS-FESOM_1",
        "IFS-NEMO_1",
        "ICON_1",
        "DGOV2",
        "IFS-FESOM_2",
        "IFS-NEMO_2",
        "ICON_2",
    ]
    df = df[new_cols]
    return df


def _to_styled_df(df):
    def get_multiindex_tuple(col):
        if "_" in col:
            return (
                "Old DGOV" if col[-1] == "1" else "New DGOV",
                col[:-2],
            )
        if col.startswith("DGOV"):
            return (
                "Old DGOV" if col[-1] == "1" else "New DGOV",
                "",
            )
        return (col, "")

    def style_color_rows(row):
        if row[("Old DGOV", "")] == "✅" and row[("New DGOV", "")] == "❌":
            return [f"background-color: {HTML_COLORS['old_dgov']}"] * len(row)
        elif row[("Old DGOV", "")] == "❌" and row[("New DGOV", "")] == "✅":
            return [f"background-color: {HTML_COLORS['new_dgov']}"] * len(row)
        else:
            return [""] * len(row)

    def replace_emojis(df):
        df["Old DGOV", ""] = (
            df["Old DGOV", ""].replace(True, "✅", inplace=False).replace(False, "❌")
        )
        df["New DGOV", ""] = (
            df["New DGOV", ""].replace(True, "✅", inplace=False).replace(False, "❌")
        )
        df = df.replace(True, "✔", inplace=False)
        df = df.replace(False, "✘", inplace=False)
        return df

    # Rename columns to remove suffixes and create multicolumns
    df.columns = MultiIndex.from_tuples(
        [get_multiindex_tuple(col) for col in df.columns]
    )
    df.replace(nan, "", inplace=True)
    df = replace_emojis(df)
    DFHTML.append(df)

    df = df.style.apply(style_color_rows, axis=1)
    DFHTML.append(df)
    return df


def _dump_html(df, output_path, title=""):
    with (
        open(output_path, "w", encoding="utf-8") as wf,
        open(HTML_FRAMEWORK, "r", encoding="utf-8") as rf,
    ):
        wf.write(
            rf.read()
            .replace("<!-- TABLE PLACEHOLDER -->", df.to_html(index=False))
            .replace("<!-- TITLE PLACEHOLDER -->", title)
        )


def dump_compared(compared):
    for kind, df in compared.items():
        df.to_csv(OUTPUT_PATH / f"compare_{kind}.csv")
        _dump_html(
            _to_styled_df(df),
            HTML_PATH / f"compare_{kind}.html",
            "Data portfolio comparison",
        )


def variable_comparison(portfolios, grib, styled=True):
    variables = {}
    for key, portfolio in portfolios.items():
        df = convert_variables(portfolio, grib)
        df = df.replace(True, "✔", inplace=False)
        df = df.replace(False, "✘", inplace=False)
        variables[key] = df

    compared = {}
    for kind in PORTFOLIO_KINDS:
        compared[kind] = compare_portfolios(variables[kind, 1], variables[kind, 2])
    return compared


def add_apps_to_compared(compared, apps):
    df = compared.copy()
    df[("APPS", "")] = df["Short name"].apply(
        lambda x: ", ".join(sorted(set(apps[apps["Variable"] == x]["Application"])))
    )
    return df[df[("APPS", "")] != ""]


def add_mother_request_to_compared(compared, apps):
    df = compared.copy()
    apps = apps.copy()
    df.rename({"Short name": "Variable"}, inplace=True)
    apps.columns = MultiIndex.from_tuples(
        [
            ("Request", col) if col != "Variable" else ("Short name", "")
            for col in apps.columns
        ]
    )
    df = merge(df, apps, how="right", on=[("Short name", "")])
    df[("ID", "")] = df[("ID", "")].fillna(-1).astype(int)
    df.fillna("-", inplace=True)
    df.sort_values(("ID", ""), inplace=True)
    # new_cols = [
    #     ("Variable", "ParamID"),
    #     ("Variable", "Name"),
    #     ("Variable", "Short Name"),
    #     ("Variable", "Units"),
    #     ("Old DGOV", ""),
    #     ("Old DGOV", "IFS-FESOM"),
    #     ("Old DGOV", "IFS-NEMO"),
    #     ("Old DGOV", "ICON"),
    #     ("New DGOV", ""),
    #     ("New DGOV", "IFS-FESOM"),
    #     ("New DGOV", "IFS-NEMO"),
    #     ("New DGOV", "ICON"),
    #     ("Request", "Application"),
    #     ("Request", "App Req. No."),
    #     ("Request", "OPA Stat"),
    #     ("Request", "OPA StatFreq"),
    # ]
    # df.rename(
    #     columns=dict(zip(df.columns, new_cols)),
    #     inplace=True
    # )
    new_cols = [
        ("ID", ""),
        ("Name", ""),
        ("Short name", ""),
        ("Units", ""),
        ("Request", "Application"),
        ("Request", "Request No."),
        ("Request", "OPA StatFreq"),
        ("Request", "OPA Stat"),
        ("Old DGOV", ""),
        ("Old DGOV", "IFS-FESOM"),
        ("Old DGOV", "IFS-NEMO"),
        ("Old DGOV", "ICON"),
        ("New DGOV", ""),
        ("New DGOV", "IFS-FESOM"),
        ("New DGOV", "IFS-NEMO"),
        ("New DGOV", "ICON"),
    ]
    df = df[new_cols]
    return df


def main():
    global apps, variables_apps, portfoilios, grib, compared, compared_apps, compared_mr

    args = _get_parser().parse_args()
    path = Path(args.path)
    apps = get_variables(path, must_match=True)
    portfolios = get_portfolios()
    grib = read_grib_params()
    dump_variables(portfolios, grib)
    compared = variable_comparison(portfolios, grib)
    dump_compared(compared)
    compared_apps = add_apps_to_compared(compared["production"], apps)
    _dump_html(
        compared_apps, HTML_PATH / "compared_apps.html", "Old-DGOV app variables"
    )
    compared_mr = add_mother_request_to_compared(compared["production"], apps)
    _dump_html(compared_mr, HTML_PATH / "compared_mr.html", "Old-DGOV MR variables")


if __name__ == "__main__":
    main()
