from argparse import ArgumentParser
from io import StringIO
from pathlib import Path
import typing as T

import requests
import pandas as pd
from bs4 import BeautifulSoup

from selenium import webdriver
from selenium.webdriver.chrome.service import Service
from selenium.webdriver.common.by import By
from selenium.webdriver.common.keys import Keys
from selenium.webdriver.support.ui import WebDriverWait
from selenium.webdriver.support import expected_conditions as EC
from webdriver_manager.chrome import ChromeDriverManager


GRIB_DB = "https://codes.ecmwf.int/grib/param-db/"
SELECTOR_OPTION = "All"
DEFAULT_GRIB_PATH = Path(__file__).resolve().parent.parent / "grib"


def _get_parser() -> ArgumentParser:
    parser = ArgumentParser()
    parser.add_argument(
        "-o",
        "--output",
        type=str,
        default=None,
    )
    return parser


def fetch_grib_params_from_webpage() -> T.Tuple[pd.DataFrame, str]:
    # Set up the WebDriver
    service = Service(ChromeDriverManager().install())
    driver = webdriver.Chrome(service=service)

    try:
        driver.get(GRIB_DB)

        # Wait until the input box is present
        input_box = WebDriverWait(driver, 10).until(
            EC.presence_of_element_located((By.ID, "n-entries-box-demo"))
        )

        # Click the input box to activate it
        input_box.click()

        # Wait until the <table> tag is present
        table_element = WebDriverWait(driver, 10).until(
            EC.presence_of_element_located((By.TAG_NAME, "table"))
        )

        # Clear the input box and enter "All"
        input_box.send_keys(Keys.CONTROL + "a")  # Use Keys.COMMAND if on Mac
        input_box.send_keys(Keys.DELETE)  # Clear the selection
        input_box.send_keys("All")

        # Simulate "down arrow" and "return" to select the first dropdown option
        input_box.send_keys(Keys.ARROW_DOWN)
        input_box.send_keys(Keys.RETURN)

        # Wait until the <table> tag is present
        WebDriverWait(driver, 10).until(
            EC.presence_of_element_located((By.TAG_NAME, "table"))
        )

        # Get the initial table HTML
        initial_table_html = table_element.get_attribute("outerHTML")

        # Wait for the table to change (i.e., be updated with new data)
        WebDriverWait(driver, 30).until(
            lambda driver: table_element.get_attribute("outerHTML")
            != initial_table_html
        )

        # Wait until the "Last updated on ..." <p> tag is present
        try:
            last_updated_text = (
                WebDriverWait(driver, 10)
                .until(
                    EC.presence_of_element_located(
                        (By.XPATH, "//p[descendant::*][contains(., 'Last updated')]")
                    )
                )
                .text
            )
        except:
            last_updated_text = None

        # Extract the page source after the table has loaded
        html = driver.page_source

    finally:
        driver.quit()

    # Parse with BeautifulSoup
    soup = BeautifulSoup(html, "html.parser")
    table = soup.find("table")  # Adjust to find your specific table

    # Use pandas to read the HTML table
    df = pd.read_html(StringIO(str(table)))[0]
    return df, last_updated_text


def dump_grib_params(
    df: pd.DataFrame,
    last_updated_text: T.Optional[str] = None,
    output_path: Path = DEFAULT_GRIB_PATH,
) -> None:

    df.to_csv(output_path / "grib.csv")

    if output_path is None:
        return

    with open(output_path / "grib_last_updated.txt", "w", encoding="utf-8") as wf:
        wf.write(last_updated_text)


def read_grib_params(
    path: Path | str = DEFAULT_GRIB_PATH,
    ret_last_updated: bool = False,
):
    path = Path(path)
    if path.is_dir():
        df = pd.read_csv(path / "grib.csv", index_col=0)
        if (path / "grib_last_updated.txt").exists():
            with open(path / "grib_last_updated.txt", "r", encoding="utf-8") as rf:
                last_updated = rf.read()
        else:
            last_updated = None
    else:
        df = pd.read_csv(path)
        last_updated = None
    if ret_last_updated:
        return df, last_updated
    return df


def main():
    args = _get_parser().parse_args()
    output = Path(args.output) if args.output is not None else DEFAULT_GRIB_PATH
    df, last_updated_text = fetch_grib_params_from_webpage()
    dump_grib_params(df, last_updated_text, output)


if __name__ == "__main__":
    main()
