#!/usr/bin/env python

"""
Process workflows repository files to extract workflow variables and generate a structured report.

steps:
    1. Collects file paths from specified directories, filtering by extensions and exclusions.
    2. Parses files to extract variable definitions into a DataFrame.
    3. Cleans the DataFrame based on ignored or non-uppercase variable names.
    4. Generates a report grouped by files or directories and writes it to a file or stdout.

examples:
    python catalog_variables.py
    python catalog_variables.py templates
    python catalog_variables.py --format md > variables.md
    python catalog_variables.py templates -f rst > variables.rst
    python catalog_variables.py templates conf -f rst -o variables.rst --upper-only --no-group
    python catalog_variables.py templates conf -e sh -f rst -U -G > variables.rst
    python catalog_variables.py templates conf -e sh py -f rst -UG -o variables.rst
"""


from argparse import ArgumentParser, RawDescriptionHelpFormatter
from collections import namedtuple
from datetime import datetime
from functools import partial
import io
import os
from pathlib import Path
import re
import subprocess
import sys
import typing as T
import warnings

import pandas as pd
from tabulate import tabulate


# Root directory of Workflows repository
ROOT_DIR = Path(__file__).resolve().parent.parent
# Mapping: Types: str to sequence of str
# Mapping: Files with given extension (key) ignore lines startng
# Mapping: with any of the given str in the sequence (value)
COMMENT_SYMBOLS = {
    ".py": ("#",),
    ".sh": ("#", "!"),
}
# Top-level directories (default)
DEFAULT_ROOTS = ["templates"]
# Variable names to ignore (default)
DEFAULT_IGNORE_VARIABLES = ["i", "j", "num"]
# Regex object with autosubmit patterb
AUTOSUBMIT_RE = re.compile(r".*(?<!\w)%(.+)%(?!\w)")
# Characters used to underline titles in rst format
RST_UNDERLINE_CHARS = ["=", "-", "+"]


def _to_fwf(df: pd.DataFrame, index: bool = True, tablefmt: str = "plain") -> str:
    """Helper function to convert DataFrame to a full-width-format table."""
    content = df.values.tolist()
    columns = list(df.columns)
    if index:
        content = [[idx] + element for idx, element in zip(df.index.tolist(), content)]
        columns = [""] + columns
    return tabulate(content, columns, tablefmt=tablefmt)


# Mapping: Types: str to Callable[DataFrame, str]
# Mapping: Allowed output formats to catalog variables script
DATAFRAME_FORMATTERS = {
    "fwf": partial(_to_fwf, index=False, tablefmt="plain"),
    "fwf-fancy": partial(_to_fwf, index=False, tablefmt="fancy_grid"),
    "rst": partial(_to_fwf, index=False, tablefmt="rst"),
    "csv": partial(pd.DataFrame.to_csv, sep="\t", index=False),
    "md": partial(pd.DataFrame.to_markdown, index=False),
    "str": partial(pd.DataFrame.to_string, index=False),
}
# Sequence: Formats that do not require a strict format. Allow header and titles
FMT_FREE_STYLE = ["fwf", "fwf-fancy", "rst", "md", "str"]

# Named tuple to store variable info
# Fields become dataframe columns
Variable = namedtuple("Variable", "Name AS Value Description Path Line")


class _PrintStream(io.TextIOBase):
    def write(self, s):
        print(s, end="")
        return len(s)


def _get_argument_parser() -> ArgumentParser:
    parser = ArgumentParser(
        description=__doc__, formatter_class=RawDescriptionHelpFormatter
    )
    parser.add_argument(
        "roots",
        type=str,
        nargs="*",
        default=DEFAULT_ROOTS,
        help="Workflow top-level directories to browse from",
    )
    parser.add_argument(
        "-e",
        "--extensions",
        type=str,
        nargs="+",
        choices=[ext.removeprefix(".") for ext in COMMENT_SYMBOLS],
        default=["sh"],
        help="Extensions from where variables are captured",
    )
    parser.add_argument(
        "-x",
        "--exclude",
        type=str,
        nargs="+",
        default=[],
        metavar="PATTERNS",
        help="Glob patterns to exclude",
    )
    parser.add_argument(
        "-i",
        "--ignore",
        type=str,
        nargs="*",
        default=DEFAULT_IGNORE_VARIABLES,
        metavar="VARIABLES",
        help="Variable names to ignore",
    )
    parser.add_argument(
        "-U",
        "--upper-only",
        action="store_true",
        help="Ignore all lower case variables",
    )
    parser.add_argument(
        "-f",
        "--format",
        type=str,
        default="str",
        choices=list(DATAFRAME_FORMATTERS),
        help="Output table format",
    )
    parser.add_argument(
        "-o",
        "--output",
        type=str,
        default=None,
        metavar="FILEPATH",
        help="Output path (if not given, print to stdout)",
    )
    parser.add_argument(
        "-F",
        "--group-by-file",
        action="store_true",
        help=(
            "If passed, group output by file. "
            "Otherwise, group output by root (each element in roots)."
        ),
    )
    parser.add_argument(
        "-G",
        "--no-group",
        action="store_false",
        help=(
            "If passed, place all output together. " "Otherwise, group output by root."
        ),
        dest="group_by_root",
    )
    return parser


def get_paths(
    root_dir: T.Union[Path, str, T.Sequence[Path], T.Sequence[str]],
    extensions: T.Optional[T.Sequence[str]] = None,
    exclude: T.Optional[T.Sequence[str]] = None,
) -> T.Sequence[Path]:
    """
    Retrieves all file paths within the specified directory or directories,
    optionally filtering by extensions or excluding specific paths.

    Args:
        root_dir: str, Path, Sequence
            The root directory or directories to search for files.
        extensions: Sequence[str], optional
            A list of file extensions to include (e.g., ["py", "txt"]).
        exclude: Sequence[str], optional
            A list of paths or substrings to exclude from the results (not yet implemented).

    Returns:
        Sequence[Path]: A list of Path objects for files matching the criteria.

    Raises:
        NotImplementedError: If the 'exclude' feature is used, as it is not yet implemented.
    """
    paths = []
    if not isinstance(root_dir, (list, tuple)):
        root_dir = [root_dir]
    for dir_ in root_dir:
        for root, _, files in os.walk(ROOT_DIR / dir_):
            for file in files:
                paths.append(Path(root) / file)
    if extensions is not None:
        extensions = list(
            set(f".{extension.removeprefix('.')}" for extension in extensions)
        )
        paths = list(filter(lambda x: x.suffix in extensions, paths))
    if exclude:
        raise NotImplementedError("'exclude' feature not implemented.")
    return paths


def _is_bash(value):
    value = value.strip()
    return bool(re.match(r"\$\{.*\}", value)) or bool(re.match(r"\$\{.*\}", value))


def _is_comment(line, extension):
    return any(line.strip().startswith(symbol) for symbol in COMMENT_SYMBOLS[extension])


def _is_condition(line):
    return line.strip().startswith("if")


def _is_export(line):
    return line.strip().startswith("export")


def _is_bad_name(line):
    # Such as bindings
    if any(line.strip().startswith(symbol) for symbol in ("--bind", "--env")):
        return True
    if re.match(r"\W", line.strip()):
        return True
    if " " in line.strip():
        return True
    return False


def process_line(line: str, path: Path, lineno: int) -> T.Union[Variable, None]:
    """
    Processes a single line from a file to extract a variable, if valid.

    Args:
        line (str): The line of text to process.
        path (Path): The file path from which the line was read.
        lineno (int): The line number in the file.

    Returns:
        Variable or None: A `Variable` object if the line contains a valid variable declaration;
                          otherwise, None.
    """
    extension = path.suffix
    comment_symbol = COMMENT_SYMBOLS[extension][0]
    if (
        line.count("=") != 1
        or _is_comment(line, extension)
        or _is_condition(line)
        or _is_bad_name(line)
    ):
        return None
    lhs, rhs = line.split("=")
    lhs = lhs.strip()
    rhs = rhs.strip()
    if comment_symbol in rhs:
        rhs, desc = rhs.split(comment_symbol, 1)
    else:
        desc = ""
    if _is_bash(rhs):
        # Potentially add a new column for variable. For now, nothing
        pass
    if _is_export(lhs):
        # Potentially add a new column for variable. For nlow, just remove "export" prefix
        lhs = lhs.removeprefix("export")
    autosumbit_match = AUTOSUBMIT_RE.match(rhs)
    if autosumbit_match:
        return Variable(
            lhs, f"%{autosumbit_match.groups()[0]}%", "", desc, str(path), lineno
        )
    return Variable(lhs, "", rhs, desc, str(path), lineno)


def _parse_file(path: Path, root: Path) -> T.List[Variable]:
    variables = []
    try:
        with open(path, "r", encoding="utf-8") as rf:
            for i, line in enumerate(rf.read().splitlines()):
                line = line.strip()
                variable = process_line(line, path.relative_to(root), i)
                if variable is None:
                    continue
                variables.append(variable)
    except UnicodeDecodeError:
        warnings.warn("Couldn't open file", path)
    return variables


def parse_files(paths: T.Sequence[Path]) -> pd.DataFrame:
    """
    Parses a list of file paths and extracts variables into a Pandas DataFrame.
    This function reads files specified in the `paths` list, processes their content
    using the `_parse_file` function, and aggregates the results into a DataFrame.

    Args:
        paths: Sequence[Path]
            A list of file paths to parse. Each file is expected
            to contain variable definitions.

    Returns:
        pd.DataFrame: A DataFrame where each row corresponds to a parsed variable
                      and contains relevant information such as the variable name,
                      value, description, and file path.
    """
    variables = []
    for path in paths:
        variables.extend(_parse_file(path, ROOT_DIR))
    return pd.DataFrame(variables)


def clean_df(df, ignore: T.Optional[T.Sequence[str]] = None, upper_only: bool = False):
    """Cleans a DataFrame by filtering out ignored or non-uppercase variable names.
    This function modifies the DataFrame in place, removing rows that match
    the provided `ignore` list or do not have uppercase-only names if `upper_only` is True.

    Args:
        df: pd.DataFrame
            The DataFrame to clean. Must contain a "Name" column, where
            each entry is the name of a variable.
        ignore: Sequence[str], optional:
            A list of variable names to exclude from the DataFrame.
        upper_only: bool, default False:
            If True, removes rows where the "Name" column does not contain
            uppercase-only variable names.
    """
    if ignore:
        df.drop(df[df["Name"].apply(lambda x: x in ignore)].index, inplace=True)
    if upper_only:
        df.drop(df[~df["Name"].apply(str.isupper)].index, inplace=True)


def _backtick(value):
    if pd.notnull(value) and value != "":
        return f"`{value}`"
    return value


def _write_group(df, stream, formatter):
    # Drop colum AS and/or Value if there are no values
    for column in ("AS", "Value"):
        if not df[column].any():
            df.drop(column, axis=1, inplace=True)
    stream.write(formatter(df.sort_values("Name", inplace=False)))
    stream.write("\n\n")


def _write_title(value: str, level: int, fmt: str, stream: T.IO):
    if fmt == "md":
        stream.write(f"{'#' * level} {fmt}\n\n")
    elif fmt == "rst":
        stream.write(f"{value}\n{RST_UNDERLINE_CHARS[level - 1] * len(value)}\n\n")
    else:
        _write_title(value, level, "rst", stream)


def _write_report_header(stream, fmt) -> str:
    if fmt not in FMT_FREE_STYLE:
        return
    date = datetime.now().date().isoformat()
    try:
        commit = subprocess.check_output(["git", "describe"], encoding="utf-8")
    except subprocess.SubprocessError:
        commit = "N/A"
    _write_title("Workflow variables", level=1, fmt=fmt, stream=stream)
    stream.write(f"Date: {date}\n\n")
    stream.write(f"Commit: {commit.strip()}\n\n")


def report_by_file(
    df: pd.DataFrame,
    stream: T.IO,
    fmt: str,
    group_first_parent: bool = True,
) -> None:
    """Generates a report by grouping variables based on file paths or their first parent directories.

    Args:
        df: pd.DataFrame
            The input DataFrame
        stream: IO
            An open, writable stream (e.g., a file or StringIO) to write the report.
        group_first_parent: bool
            If True, groups variables by the first component of the "Path" column.
            If False, groups directly by the "Path" column.

    Returns:
        None: Writes the report to the provided stream.
    """
    df_formatter = DATAFRAME_FORMATTERS[fmt]
    _write_report_header(stream, fmt)

    for column in ("Name", "AS", "Value"):
        df[column] = df[column].apply(_backtick)
    if group_first_parent:
        df["FirstPathComponent"] = df["Path"].apply(
            lambda x: Path(x).parts[0].capitalize()
        )
        for first_parent, parent_group in df.groupby("FirstPathComponent"):
            _write_title(str(first_parent), 2, fmt, stream)
            for path, path_df in parent_group.groupby("Path"):
                _write_title(str(path), 3, fmt, stream)
                path_df.drop("FirstPathComponent", axis=1, inplace=True)
                path_df.drop("Path", axis=1, inplace=True)
                _write_group(path_df, stream, df_formatter)
    else:
        for path, path_df in df.groupby("Path"):
            _write_title(str(path), 2, fmt, stream)
            path_df.drop("Path", axis=1, inplace=True)
            _write_group(path_df, stream, df_formatter)


def _longest_common_path(paths):
    if not paths:
        return None
    path_objects = [Path(p) for p in paths]
    common_parts = path_objects[0].parts
    for path in path_objects[1:]:
        common_parts = tuple(
            part for part, other in zip(common_parts, path.parts) if part == other
        )
        if not common_parts:
            break
    return str(Path(*common_parts))


def _compile_path_in_duplicates(df, subset):
    duplicate_groups = df.duplicated(subset=subset, keep=False)
    if duplicate_groups.any():
        grouped = df.groupby(list(subset))
        for _, group in grouped:
            if len(group) > 1:
                common_path = _longest_common_path(group["Path"].tolist())
                is_dir = (ROOT_DIR / common_path).is_dir()
                new_path = f"{common_path}{'/*' if is_dir else ''} ({len(group)})"
                df.loc[group.index, "Path"] = new_path
                df.loc[group.index, "Line"] = None
    df.drop_duplicates(subset=subset, inplace=True)
    if df["Line"].isna().any():
        df["Line"] = df["Line"].map(lambda x: "-" if pd.isna(x) else str(int(x)))


def _get_select_fun(prefix: str):
    def fun(x: str):
        return x.startswith(prefix)

    return fun


def report_by_group(
    df: pd.DataFrame,
    stream: T.IO,
    fmt: str,
    restrict: T.Optional[T.Union[str, T.Sequence[str]]] = None,
) -> None:
    """Generates a report by grouping variables based on a restricted set of paths.

    Args:
        df: pd.DataFrame
            The input DataFrame
        stream: IO
            An open, writable stream (e.g., a file or StringIO) to write the report.
        restrict: str, Sequence[str], optional
            A target path prefix (or list) to filter the variables.
            If not passed, writes all variables.

    Returns:
        None: Writes the report to the provided stream.
    """
    df_formatter = DATAFRAME_FORMATTERS[fmt]
    _write_report_header(stream, fmt)

    for column in ("Name", "AS", "Value"):
        df[column] = df[column].apply(_backtick)
    if not isinstance(restrict, (list, tuple)):
        restrict = [restrict]
    for target_path in restrict:
        if target_path is not None:
            target_path_title = (
                target_path.capitalize() if "/" not in target_path else target_path
            )
            _write_title(target_path_title, 2, fmt, stream)
            target_df = df.copy()[df["Path"].apply(_get_select_fun(target_path))]
        else:
            target_df = df
        _compile_path_in_duplicates(target_df, ["Name", "AS", "Value"])
        _write_group(target_df, stream, df_formatter)


def main():
    """Main."""
    args = _get_argument_parser().parse_args(sys.argv[1:])
    paths = get_paths(args.roots, args.extensions, args.exclude)
    df = parse_files(paths)
    clean_df(df, args.ignore, args.upper_only)
    stream = open(args.output, "w", encoding="utf-8") if args.output else _PrintStream()
    if args.group_by_file:
        report_by_file(df, stream, args.format, group_first_parent=True)
    elif args.group_by_root and args.roots:
        if args.format in FMT_FREE_STYLE:
            report_by_group(df, stream, args.format, args.roots)
        else:
            if len(args.roots) > 1:
                warnings.warn(
                    f"Format '{args.format}' does not allow grouping, "
                    f"so running with option --no-group."
                )
            report_by_group(df, stream, args.format)
    else:
        report_by_group(df, stream, args.format)
    stream.close()


if __name__ == "__main__":
    main()
